// ABBYY® Mobile Capture © 2019 ABBYY Production LLC.
// ABBYY is a registered trademark or a trademark of ABBYY Software Ltd.

import AbbyyUI

class ImageCollectionViewCell: UICollectionViewCell {
	var imageView: UIImageView? {
		willSet {
			imageView?.removeFromSuperview()
		}
		didSet {
			if let imageView = imageView {
				imageView.contentMode = .scaleAspectFill
				contentView.addSubview(imageView)
			}
		}
	}

	override func prepareForReuse()
	{
		super.prepareForReuse()
		imageView = nil
	}
}

class StartViewController: UIViewController {

	let dataStorage = DataStorage()
	@IBOutlet weak var addPagesButton: UIButton!
	@IBOutlet weak var placeholderImageView: UIImageView!
	@IBOutlet weak var versionLabel: UILabel!
	@IBOutlet weak var sharePdfButton: UIButton!

	fileprivate let minPointsPerItem = 88

	var columnsCount: Int {
		return Int(UIScreen.main.bounds.width) / minPointsPerItem
	}

	@IBOutlet weak var pagesCollection: UICollectionView!
	var collectionCellsInset: CGFloat = 2

	var captureController: CustomCaptureViewController?
	var capturedCountOnCurrentSession = 0
}

// MARK: UIViewController overrides & UI

extension StartViewController {
	override func viewDidLoad()
	{
		super.viewDidLoad()
		pagesCollection.delegate = self
		pagesCollection.dataSource = self
		versionLabel.text = "Build Number: \(RecognitionEngine.version)"
	}

	override func viewWillAppear(_ animated: Bool)
	{
		super.viewWillAppear(animated)
		let hasCapturedFrames = dataStorage.itemsCount > 0
		addPagesButton.isHidden = !hasCapturedFrames
		sharePdfButton.isHidden = !hasCapturedFrames
		placeholderImageView.isHidden = hasCapturedFrames
		pagesCollection.reloadData()
	}

	override var supportedInterfaceOrientations: UIInterfaceOrientationMask {
		return .portrait
	}
}

// MARK: Actions

extension StartViewController {

	func presentCaptureController()
	{
		guard let engine = RecognitionEngine.shared else {
			let alert = UIAlertController(
				title: "SomethingWentWrong".localized,
				message: "InvalidLicenseMessage".localized,
				preferredStyle: .alert)
			alert.addAction(UIAlertAction(title: "Ok".localized, style: .default))
			UIViewController.top?.present(alert, animated: true)
			return;
		}
		// Capture Controller Configuration
		// Create AUICaptureController subclass
		let sb = UIStoryboard(name: "\(CustomCaptureViewController.self)", bundle: nil)
		captureController = sb.instantiateInitialViewController() as? CustomCaptureViewController
		
		guard let captureController = captureController else {
			return
		}

		// Create capture scenario using RTREngine
		let captureScenario = AUIImageCaptureScenario(engine: engine)

		// Configure capture scenario
		// Image will be cropped and quadrangle result will be nil
		captureScenario.cropEnabled = true

		// Send recognition callback to self
		// See "AUIImageCaptureScenarioDelegate implementation" section
		captureScenario.delegate = self

		// Bind capture scenario with capture controller
		captureController.captureScenario = captureScenario

		captureController.framesCaptured = dataStorage.itemsCount

		if dataStorage.itemsCount > 0 {
			let lastIndexPath = IndexPath(row: dataStorage.itemsCount - 1, section: 0)
			let cell = pagesCollection.cellForItem(at: lastIndexPath) as? ImageCollectionViewCell
			// if last image is not loaded, load it synchronously from url
			if let image = cell?.imageView?.image {
				captureController.imageMiniature = image
			} else if
				let imageUrl = dataStorage.content.last,
				let imageData = try? Data(contentsOf: imageUrl) {
				captureController.imageMiniature = UIImage(data: imageData)
			}
		}
		capturedCountOnCurrentSession = 0
		captureController.onThumbnailViewTapped = { [unowned self] in
			self.captureController?.dismiss(animated: true)
		}
		present(captureController, animated: true)
	}

	@IBAction func onScanNewDocumentTapped(sender: UIButton)
	{
		dataStorage.clear()
		presentCaptureController()
	}

	@IBAction func onAddPagesTapped(sender: UIButton)
	{
		presentCaptureController()
	}

	@IBAction func onSharePdfTapped(sender: UIButton)
	{
		weak var weakSelf = self
		func prettyFilename() -> String
		{
			let currentDate = Date()
			let formatter = DateFormatter()
			formatter.dateStyle = .medium
			formatter.timeStyle = .medium
			let dateString = formatter.string(from: currentDate)
			return "ImageCapture - \(dateString).pdf"
		}

		func showShareController(with url: URL)
		{
			let activityViewController = UIActivityViewController(
				activityItems: [url],
				applicationActivities: nil)
			if let popover = activityViewController.popoverPresentationController {
				activityViewController.modalPresentationStyle = .popover
				popover.sourceRect = sharePdfButton.bounds
				popover.permittedArrowDirections = [.down, .up]
				popover.sourceView = sharePdfButton
			}
			weakSelf?.present(activityViewController, animated: true, completion: nil)
		}

		func showExportErrorWarning(with error: Error)
		{
			let alert = UIAlertController(
				title: "SomethingWentWrong".localized,
				message: String.localizedStringWithFormat(
					"PdfExportWithError".localized, "\(error)"),
				preferredStyle: .alert)
			alert.addAction(UIAlertAction(title: "Ok".localized, style: .default))
			UIViewController.top?.present(alert, animated: true)
		}

		let pdfFilename = prettyFilename()
		let progressAlert = UIAlertController(
			title: "ExportInProgress".localized,
			message: "",
			preferredStyle: .alert)
		let exportingProgressCallack: (ImageExporter.PdfExportStatus) -> () = { status in
			DispatchQueue.main.async {
				progressAlert.message = "\(status.currentAction.rawValue.capitalized): \(status.pagesProcessed) / \(status.pagesCount).."
			}
		}

		self.present(progressAlert, animated: true)

		DispatchQueue.global(qos: .default).async { [weak self] in
			do {
				try self?.dataStorage.exportAsPdf(
					filename: pdfFilename,
					progress: exportingProgressCallack)
				DispatchQueue.main.async {
					let pdfFilePath = self?.dataStorage
						.exporter
						.tmpDirectory
						.appendingPathComponent(pdfFilename)
					progressAlert.dismiss(animated: true, completion: {
						if let filepath = pdfFilePath {
							showShareController(with: filepath)
						} else {
							showExportErrorWarning(with: ImageExporter.ExportError.internalError)
						}
					})
				}
			} catch {
				DispatchQueue.main.async {
					self?.dismiss(animated: true, completion: {
						showExportErrorWarning(with: error)
					})
				}
			}
		}
	}
}

// MARK: AUIImageCaptureScenarioDelegate implementation

extension StartViewController: AUIImageCaptureScenarioDelegate & UINavigationControllerDelegate
{
	func captureScenario(_
		scenario: AUICaptureScenario,
		didFailWithError error: Error)
	{
		let alert = UIAlertController(
			title: "SomethingWentWrong".localized,
			message: error.localizedDescription,
			preferredStyle: .alert)
		alert.addAction(UIAlertAction(
			title: "Ok".localized,
			style: .cancel))
		UIViewController.top?.present(alert, animated: true, completion: { [weak self] in
			self?.captureController?.paused = true
		})
	}

	func captureScenario(_
		captureScenario: AUIImageCaptureScenario,
		didCaptureImageWith result: AUIImageCaptureResult)
	{
		let sb = UIStoryboard(name: "\(SingleImagePreviewController.self)", bundle: nil)
		if let vc = sb.instantiateInitialViewController() as? SingleImagePreviewController {
			vc.image = result.image
			vc.pageNumber = dataStorage.itemsCount + 1
			vc.delegate = self
			captureController?.present(UINavigationController(rootViewController: vc), animated: true)
		}
	}

	func captureScenarioDidCancel(_ scenario: AUICaptureScenario) {
		let onCancelledByUser: (UIViewController?) -> () = { [unowned self] cvc in
			self.dataStorage.removeLast(count: self.capturedCountOnCurrentSession)
			cvc?.dismiss(animated: true)
		}
		guard capturedCountOnCurrentSession > 0 else {
			onCancelledByUser(captureController)
			return
		}
		captureController?.setPaused(true, animated: true)

		let onConfirmAlert = UIAlertAction(
			title: "Confirm".localized,
			style: .destructive) { [weak self] _ in
				onCancelledByUser(self?.captureController)
		}
		let onCancelAlert = UIAlertAction(
			title: "Cancel".localized,
			style: .cancel) { [weak self] _ in
				self?.captureController?.closeButton?.isEnabled = true
				self?.captureController?.setPaused(false, animated: true)
		}
		let alert = UIAlertController(
			title: "DectructiveActionWarning".localized,
			message: "AllPagesOnCurrentSessionWillBeDeletedWarning".localized,
			preferredStyle: .alert)
		alert.addAction(onCancelAlert)
		alert.addAction(onConfirmAlert)
		UIViewController.top?.present(alert, animated: true)
	}
}

// MARK: SingleImagePreviewControllerDelegate implementation

extension StartViewController: SingleImagePreviewControllerDelegate
{
	func addPageFrom(
		previewController: SingleImagePreviewController,
		onFailure: (Error?) -> (),
		onSuccess: () -> ())
	{
		guard let image = previewController.image else {
			onFailure(.none)
			return
		}
		do {
			try dataStorage.save(image: image)
			captureController?.framesCaptured = dataStorage.itemsCount
			captureController?.imageMiniature = image
			onSuccess()
		} catch {
			onFailure(error)
		}
	}

	func onAddPage(
		according previewController: SingleImagePreviewController,
		sender: Any?)
	{
		addPageFrom(previewController: previewController, onFailure: { [weak self] error in
			let action = UIAlertAction(title: "Ok".localized, style: .cancel) { _ in
				self?.captureController?.dismiss(animated: true)
			}
			let alert = UIAlertController(
				title: "SomethingWentWrong".localized,
				message: String.localizedStringWithFormat(
					"SavePageWithError".localized, String(describing: error)),
				preferredStyle: .alert)
			alert.addAction(action)
			UIViewController.top?.present(alert, animated: true)
		}, onSuccess: { [weak self] in
			self?.capturedCountOnCurrentSession = capturedCountOnCurrentSession.advanced(by: 1)
			self?.captureController?.dismiss(animated: true)
		})
	}

	func onRejectPage(
		according previewController: SingleImagePreviewController,
		sender: Any?)
	{
		captureController?.dismiss(animated: true)
	}

	func onAllPagesCaptured(
		according previewController: SingleImagePreviewController,
		sender: Any?)
	{
		addPageFrom(previewController: previewController, onFailure: { [weak self] error in
			let action = UIAlertAction(title: "Ok".localized, style: .cancel) { _ in
				self?.captureController?.dismiss(animated: true)
			}
			let alert = UIAlertController(
				title: "SomethingWentWrong".localized,
				message: String.localizedStringWithFormat(
					"SavePageWithError".localized, String(describing: error)),
				preferredStyle: .alert)
			alert.addAction(action)
			UIViewController.top?.present(alert, animated: true)
		}, onSuccess: { [weak self] in
			self?.dismiss(animated: true)
		})
	}
}

// MARK: UICollectionViewDelegate implementation

extension StartViewController: UICollectionViewDelegate & UICollectionViewDelegateFlowLayout
{
	func collectionView(_ collectionView: UICollectionView,
		layout collectionViewLayout: UICollectionViewLayout,
		sizeForItemAt indexPath: IndexPath) -> CGSize
	{
		return CGSize(
			width: view.frame.width / CGFloat(columnsCount) - collectionCellsInset * 2,
			height: view.frame.width / CGFloat(columnsCount) - collectionCellsInset * 2);
	}

	func collectionView(_ collectionView: UICollectionView,
		layout collectionViewLayout: UICollectionViewLayout,
		minimumLineSpacingForSectionAt section: Int) -> CGFloat
	{
		return collectionCellsInset
	}

	func collectionView(_ collectionView: UICollectionView,
		layout collectionViewLayout: UICollectionViewLayout,
		minimumInteritemSpacingForSectionAt section: Int) -> CGFloat
	{
		return collectionCellsInset
	}

	func collectionView(_ collectionView: UICollectionView,
			layout collectionViewLayout: UICollectionViewLayout,
			insetForSectionAt section: Int) -> UIEdgeInsets
	{
		return UIEdgeInsets(
			top: collectionCellsInset,
			left: collectionCellsInset,
			bottom: collectionCellsInset,
			right: collectionCellsInset);
	}

	func collectionView(_ collectionView: UICollectionView,
		willDisplay cell: UICollectionViewCell,
		forItemAt indexPath: IndexPath)
	{
		guard let imageCell = cell as? ImageCollectionViewCell else { return }

		let updateCellClosure: (UIImage?) -> Void = { [weak self] image in
			imageCell.imageView?.image = image
			self?.dataStorage.loadingOperations.removeValue(forKey: indexPath.row)
		}
		if let dataLoader = dataStorage.loadingOperations[indexPath.row] {
			if let image = dataLoader.image {
				imageCell.imageView?.image = image
				dataStorage.loadingOperations.removeValue(forKey: indexPath.row)
			} else {
				dataLoader.loadingCompleteHandler = updateCellClosure
			}
		} else {
			if let dataLoader = dataStorage.fetch(at: indexPath.row) {
				dataLoader.loadingCompleteHandler = updateCellClosure
				dataStorage.loadingQueue.addOperation(dataLoader)
				dataStorage.loadingOperations[indexPath.row] = dataLoader
			}
		}
	}

	func collectionView(_ collectionView: UICollectionView,
		didEndDisplaying cell: UICollectionViewCell,
		forItemAt indexPath: IndexPath)
	{
		if let dataLoader = dataStorage.loadingOperations[indexPath.row] {
			dataLoader.cancel()
			dataStorage.loadingOperations.removeValue(forKey: indexPath.row)
		}
	}
}

// MARK: UICollectionViewDataSource implementation

extension StartViewController: UICollectionViewDataSource
{
	func numberOfSections(in collectionView: UICollectionView) -> Int
	{
		return dataStorage.itemsCount > 0 ? 1 : 0
	}

	func collectionView(_ collectionView: UICollectionView, numberOfItemsInSection section: Int) -> Int
	{
		return dataStorage.itemsCount
	}

	func collectionView(_ collectionView: UICollectionView, cellForItemAt indexPath: IndexPath) -> UICollectionViewCell
	{
		let cell = collectionView.dequeueReusableCell(
			withReuseIdentifier: "PageCell",
			for: indexPath) as? ImageCollectionViewCell
		
		let cellSize = CGSize(
			width: view.frame.width / CGFloat(columnsCount),
			height: view.frame.width / CGFloat(columnsCount))
		
		cell?.imageView = UIImageView(frame: CGRect(origin: .zero, size: cellSize))
		
		return cell ?? UICollectionViewCell()
	}
}
