// ABBYY® Mobile Capture © 2019 ABBYY Production LLC.
// ABBYY is a registered trademark or a trademark of ABBYY Software Ltd.

#import "DocumentManager.h"

/// Base functionality for files containers with various types
@interface FileContainer : NSObject

/// Returns root directory path
@property (nonatomic, strong) NSString* directory;
/// File Name of controlled files
@property (nonatomic, readonly) NSString* filename;
/// Creates container with custom storage path
- (instancetype)initWithDirectory:(NSString*)directory;
/// Add images to container
- (NSString*)addImages:(NSArray<UIImage*>*)images;
/// Add image to container
- (NSString*)addImage:(UIImage*)image;
/// Remove all files from container
- (void)clear;

@end

@implementation FileContainer

- (instancetype)initWithDirectory:(NSString*)directory
{
	self = [super init];
	if(self != nil) {
		self.directory = directory;
	}
	return self;
}

- (NSString*)addImages:(NSArray<UIImage*>*)images
{
	if(images.count == 0) {
		return nil;
	}
	// Generate container filename from current time
	NSString* path = [self.directory stringByAppendingPathComponent:self.filename];
	// Create export operation instance with file output stream
	RTRFileOutputStream* stream = [[RTRFileOutputStream alloc] initWithFilePath:path];
	id<RTRCoreAPIExportOperation> operation = [self exportOperation:stream];
	// Add all images to container
	for(UIImage* image in images) {
		[operation addPageWithImage:image];
	}
	// operation has to be closed
	return [operation close] ? path : nil;
}

- (NSString*)addImage:(UIImage*)image
{
	return [self addImages:@[image]];
}

- (void)clear
{
	// Remove all files at root directory
	[[NSFileManager defaultManager] removeItemAtPath:self.directory error:nil];
}

- (NSString*)directory
{
	if(_directory != nil) {
		return _directory;
	}
	// Temp folder
	return NSTemporaryDirectory();
}

- (NSString*)filename
{
	// override
	return nil;
}

- (id<RTRCoreAPIExportOperation>)exportOperation:(RTRFileOutputStream*)stream
{
	// override
	return nil;
}

- (id<RTRCoreAPI>)coreApi
{
	return [[RecognizerEngine rtrEngine] createCoreAPI];
}

@end

//----------------------------------------------------------

@interface ImageContainer : FileContainer
/// Saved images
@property (nonatomic, readonly) NSArray<NSString*>* imagePaths;

@end

@implementation ImageContainer

- (NSString*)addImages:(NSArray<UIImage*>*)images
{
	if(images.count == 1) {
		return [super addImages:images];
	}
	for(UIImage* image in images) {
		[super addImages:@[image]];
	}
	return self.directory;
}

- (id<RTRCoreAPIExportOperation>)exportOperation:(RTRFileOutputStream*)stream
{
	return [self.coreApi createExportToJpgOperation:stream];
}

- (NSString*)fileExtension
{
	return @"jpg";
}

- (NSString*)filename
{
	return [NSString stringWithFormat:@"%ld.%@", (NSInteger)[NSDate date].timeIntervalSince1970 * 1000, self.fileExtension];
}

- (NSArray<NSString*>*)imagePaths
{
	NSFileManager* fileManager = [NSFileManager defaultManager];
	// Get all files in root directory
	NSArray* files = [fileManager contentsOfDirectoryAtPath:self.directory error:nil];
	// Exclude extra
	NSString* format = [NSString stringWithFormat:@"self ENDSWITH '.%@'", self.fileExtension];
	files = [files filteredArrayUsingPredicate:[NSPredicate predicateWithFormat:format]];
	// Sort by name (the same as by creation date in this case)
	files = [files sortedArrayUsingSelector:@selector(localizedCaseInsensitiveCompare:)];
	NSMutableArray* paths = [NSMutableArray arrayWithCapacity:files.count];
	for(NSString* file in files) {
		NSString* path = [self.directory stringByAppendingPathComponent:file];
		[paths addObject:path];
	}
	return paths;
}

@end

//----------------------------------------------------------

@interface PdfContainer : FileContainer

- (NSString*)generatePdfFrom:(NSArray<UIImage*>*)images;

@end

@implementation PdfContainer

- (id<RTRCoreAPIExportOperation>)exportOperation:(RTRFileOutputStream*)stream
{
	return [self.coreApi createExportToPdfOperation:stream];
}

- (NSString*)filename
{
	NSDateFormatter* formatter = [NSDateFormatter new];
	formatter.dateStyle = NSDateFormatterMediumStyle;
	formatter.timeStyle = NSDateFormatterMediumStyle;
	NSString* dateString = [formatter stringFromDate:[NSDate date]];
	return [NSString stringWithFormat:@"ImageCapture - %@.pdf", dateString];
}

- (NSString*)generatePdfFrom:(NSArray<UIImage*>*)images
{
	return [self addImages:images];
}

@end

//----------------------------------------------------------

@interface DocumentManager ()

@property (nonatomic, strong) ImageContainer* imageContainer;
@property (nonatomic, strong) PdfContainer* pdfContainer;

@end

@implementation DocumentManager

+ (instancetype)defaultManager
{
	DocumentManager* manager = [[self class] new];
	// Document folder
	NSString* directory = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES).firstObject;
	manager.imageContainer = [[ImageContainer alloc] initWithDirectory:directory];
	return manager;
}

+ (instancetype)tempSessionManager
{
	return [[self class] new];
}

- (ImageContainer*)imageContainer
{
	if(_imageContainer == nil) {
		_imageContainer = [ImageContainer new];
	}
	return _imageContainer;
}

- (PdfContainer*)pdfContainer
{
	if(_pdfContainer == nil) {
		_pdfContainer = [PdfContainer new];
	}
	return _pdfContainer;
}

- (NSArray<NSString*>*)imagePaths
{
	return self.imageContainer.imagePaths;
}

- (NSString*)saveImage:(UIImage*)image
{
	return [self.imageContainer addImage:image];
}

- (NSArray<NSString*>*)saveImageFiles:(NSArray<NSString*>*)pathArray
{
	NSMutableArray* results = [NSMutableArray arrayWithCapacity:pathArray.count];
	for(NSString* filePath in pathArray) {
		NSError* error = nil;
		NSString* destination = [self.imageContainer.directory stringByAppendingPathComponent:filePath.lastPathComponent];
		if([[NSFileManager defaultManager] copyItemAtPath:filePath toPath:destination error:&error]) {
			[results addObject:destination];
		}
	}
	NSAssert(results.count == pathArray.count, @"Copy files failed");
	return results;
}

- (void)generatePdfWithCompletion:(PdfCompletion)completion
{
	// remove old files
	[self.pdfContainer clear];
	// generate new
	dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_HIGH, 0), ^{
		NSArray* pathArray = self.imageContainer.imagePaths;
		NSMutableArray* images = [NSMutableArray arrayWithCapacity:pathArray.count];
		for(NSString* path in pathArray) {
			UIImage* image = [UIImage imageWithContentsOfFile:path];
			if(image != nil) {
				[images addObject:image];
			}
		}
		NSString* path = [self.pdfContainer generatePdfFrom:images];
		dispatch_async(dispatch_get_main_queue(), ^{
			if(completion != nil) {
				completion(path);
			}
		});
	});
}

- (void)removeAllFiles
{
	[self.imageContainer clear];
	[self.pdfContainer clear];
}

- (void)removeFileAt:(NSString*)path
{
	[[NSFileManager defaultManager] removeItemAtPath:path error:nil];
}

@end
