// ABBYY® Mobile Capture © 2019 ABBYY Production LLC.
// ABBYY is a registered trademark or a trademark of ABBYY Software Ltd.

#import "ViewController.h"
#import "PreviewViewController.h"
#import "PreviewButton.h"
#import "DocumentManager.h"
#import "ImageLoader.h"
#import <AbbyyUI/AbbyyUI.h>

/// Capture controller overlay view to add preview button
@interface OverlayView : UIView
/// Preview button with badge count
@property (nonatomic, weak) IBOutlet PreviewButton* previewButton;

@end

@implementation OverlayView

@end

/// Captured document thumbnails
@interface CollectionViewCell : UICollectionViewCell

@property (nonatomic, weak) IBOutlet UIImageView* imageView;

@end

@implementation CollectionViewCell

- (void)prepareForReuse
{
	[super prepareForReuse];
	self.imageView.image = nil;
}

@end

@interface ViewController () <AUIImageCaptureScenarioDelegate, PreviewViewControllerDelegate, UICollectionViewDelegate, UICollectionViewDataSource, UICollectionViewDelegateFlowLayout>

/// Camera controller
@property (nonatomic, strong) AUICaptureController* captureController;
/// Document handler
@property (nonatomic, strong) DocumentManager* documentManager;
/// Temporary session document handler
@property (nonatomic, strong) DocumentManager* tempSessionDocumentManager;
/// Captured images paths
@property (nonatomic, strong) NSMutableArray<NSString*>* imagePaths;
/// Captured images paths
@property (nonatomic, strong) NSMutableArray<NSString*>* tempImagePaths;

/// Thumbnails collection
@property (nonatomic, weak) IBOutlet UICollectionView* collectionView;
@property (nonatomic, strong) ImageLoader* imageLoader;

/// Add more pages button
@property (nonatomic, weak) IBOutlet UIButton* addButton;
/// Share PDF button
@property (nonatomic, weak) IBOutlet UIButton* shareButton;
/// App build version text
@property (nonatomic, weak) IBOutlet UILabel* versionLabel;
@property (nonatomic, strong) IBOutlet OverlayView* overlayView;

@end

@implementation ViewController

- (void)viewDidLoad
{
	[super viewDidLoad];
	// disable buttons if nothing to add and share
	self.addButton.hidden = self.imagePaths.count == 0;
	self.shareButton.hidden = self.imagePaths.count == 0;
	self.versionLabel.text = [NSString stringWithFormat:@"Build Number: %@", [RecognizerEngine version]];
	// check temp files in case when previous session was interrupted
	[self saveTempFiles];
}

- (UIInterfaceOrientationMask)supportedInterfaceOrientations
{
	return UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPhone ? UIInterfaceOrientationMaskPortrait : UIInterfaceOrientationMaskAll;
}

- (DocumentManager*)documentManager
{
	if(_documentManager == nil) {
		_documentManager = [DocumentManager defaultManager];
	}
	return _documentManager;
}

- (DocumentManager*)tempSessionDocumentManager
{
	if(_tempSessionDocumentManager == nil) {
		_tempSessionDocumentManager = [DocumentManager tempSessionManager];
	}
	return _tempSessionDocumentManager;
}

- (NSMutableArray<NSString*>*)imagePaths
{
	if(_imagePaths == nil) {
		_imagePaths = [self.documentManager.imagePaths mutableCopy];
	}
	return _imagePaths;
}

- (NSMutableArray<NSString*>*)tempImagePaths
{
	if(_tempImagePaths == nil) {
		_tempImagePaths = [self.tempSessionDocumentManager.imagePaths mutableCopy];
	}
	return _tempImagePaths;
}

- (ImageLoader*)imageLoader
{
	if(_imageLoader == nil) {
		_imageLoader = [ImageLoader new];
	}
	return _imageLoader;
}

- (void)saveImage:(UIImage*)image temporary:(BOOL)temporary
{
	if(temporary) {
		NSString* filePath = [self.tempSessionDocumentManager saveImage:image];
		NSLog(@"Image temporary saved to: %@", filePath);
		[self.tempImagePaths addObject:filePath];
	} else {
		NSString* filePath = [self.documentManager saveImage:image];
		NSLog(@"Image saved to: %@", filePath);
		[self.imagePaths addObject:filePath];
	}
	// Show preview button with corresponding badge count on camera view
	OverlayView* overlayView = (id)self.captureController.cameraOverlayView;
	overlayView.previewButton.image = image;
	overlayView.previewButton.badgeNumber = self.imagePaths.count + self.tempImagePaths.count;
	overlayView.hidden = NO;
}

- (void)saveTempFiles
{
	// save all temp images
	NSArray* array = [self.documentManager saveImageFiles:self.tempImagePaths];
	[self.imagePaths addObjectsFromArray:array];
	// clear temp images
	self.tempImagePaths = nil;
	[self.tempSessionDocumentManager removeAllFiles];
}

- (void)completeCaptureSessionWithImage:(UIImage*)image
{
	[self saveTempFiles];
	if(image != nil) {
		[self saveImage:image temporary:NO];
	}
	self.addButton.hidden = self.imagePaths.count == 0;
	self.shareButton.hidden = self.imagePaths.count == 0;
	[self hideCaptureController];
}

- (void)showCaptureController
{
	if(RecognizerEngine.rtrEngine == nil) {
		[self showInvalidEngineMessage];
		return;
	}
	AUICaptureController* captureController = [AUICaptureController new];

	PreviewButton* previewButton = self.overlayView.previewButton;
	[previewButton addTarget:self action:@selector(didPressPreview:) forControlEvents:UIControlEventTouchUpInside];
	previewButton.image = self.imagePaths.count > 0 ? [UIImage imageWithContentsOfFile:self.imagePaths.lastObject] : nil;
	previewButton.badgeNumber = self.imagePaths.count;
	
	captureController.cameraOverlayView = self.overlayView;
	captureController.cameraOverlayView.hidden = self.imagePaths.count == 0;
	// Scenario contains documents capture settings. Set it up before camera view presenting
	AUIImageCaptureScenario* scenario = [[AUIImageCaptureScenario alloc] initWithEngine:[RecognizerEngine rtrEngine]];
	scenario.delegate = self;
	// To get UIImage in results
	scenario.cropEnabled = YES;
	captureController.captureScenario = scenario;
	captureController.flashButton.hidden = ![UIImagePickerController isFlashAvailableForCameraDevice:UIImagePickerControllerCameraDeviceRear];
	[self presentViewController:captureController animated:YES completion:nil];
	self.captureController = captureController;
}

- (void)showInvalidEngineMessage
{
	UIAlertController* alert = [UIAlertController
		alertControllerWithTitle: NSLocalizedString(@"SomethingWentWrong", nil)
		message: NSLocalizedString(@"InvalidLicenseMessage", nil)
		preferredStyle:UIAlertControllerStyleAlert];
	[alert addAction:[UIAlertAction actionWithTitle:NSLocalizedString(@"Ok", nil) style:UIAlertActionStyleCancel handler:nil]];
	[self presentViewController:alert animated:YES completion:nil];
}

- (void)hideCaptureController
{
	[self.collectionView reloadData];
	[self dismissViewControllerAnimated:YES completion:nil];
}

- (void)didPressPreview:(id)sender
{
	[self completeCaptureSessionWithImage:nil];
}

- (IBAction)didPressStartButton:(UIButton*)sender
{
	// Remove previous results
	[self.imageLoader clear];
	[self.documentManager removeAllFiles];
	self.imagePaths = nil;
	[self.collectionView reloadData];
	self.addButton.hidden = YES;
	self.shareButton.hidden = YES;
	[self showCaptureController];
}

- (IBAction)didPressAddButton:(UIButton*)sender
{
	[self showCaptureController];
}

- (IBAction)didPressExportButton:(UIButton*)sender
{
	UIAlertController* progress = [UIAlertController alertControllerWithTitle:NSLocalizedString(@"ExportInProgress", nil) message:nil preferredStyle:UIAlertControllerStyleAlert];
	[self presentViewController:progress animated:YES completion:nil];
	// Share PDF file from captured images
	[self.documentManager generatePdfWithCompletion:^(NSString* path) {
		if(path.length == 0) {
			NSLog(@"Nothing to share");
			[self dismissViewControllerAnimated:YES completion:^{
				[self showInvalidEngineMessage];
			}];
			return;
		}
		UIActivityViewController* viewController = [[UIActivityViewController alloc] initWithActivityItems:@[[NSURL fileURLWithPath:path]] applicationActivities:nil];
		viewController.popoverPresentationController.sourceView = self.shareButton;
		viewController.popoverPresentationController.sourceRect = self.shareButton.bounds;
		DocumentManager* documentManager = self.documentManager;
		viewController.completionWithItemsHandler = ^(UIActivityType activityType, BOOL completed, NSArray* returnedItems, NSError* activityError) {
			if(completed) {
				NSLog(@"PDF sharing completed");
				[documentManager removeFileAt:path];
			}
		};
		[self dismissViewControllerAnimated:YES completion:^{
			[self presentViewController:viewController animated:YES completion:nil];
		}];
	}];
}

#pragma mark - Abbyy Mobile UI

- (void)captureScenarioDidCancel:(AUICaptureScenario*)scenario
{
	if(self.tempImagePaths.count > 0) {
		UIAlertController* confirmation = [UIAlertController alertControllerWithTitle:NSLocalizedString(@"DectructiveActionWarning", nil) message:NSLocalizedString(@"AllPagesOnCurrentSessionWillBeDeletedWarning", nil) preferredStyle:UIAlertControllerStyleAlert];
		UIAlertAction* cancel = [UIAlertAction actionWithTitle:NSLocalizedString(@"Cancel", nil) style:UIAlertActionStyleCancel handler:^(UIAlertAction* action) {
			self.captureController.closeButton.enabled = YES;
			[self.captureController setPaused:NO animated:YES];
		}];
		UIAlertAction* ok = [UIAlertAction actionWithTitle:NSLocalizedString(@"Confirm", nil) style:UIAlertActionStyleDestructive handler:^(UIAlertAction* action) {
			[self.captureController dismissViewControllerAnimated:YES completion:^{
				dispatch_async(dispatch_get_main_queue(), ^{
					self.tempImagePaths = nil;
					[self.tempSessionDocumentManager removeAllFiles];
					[self hideCaptureController];
				});
			}];
		}];
		[confirmation addAction:ok];
		[confirmation addAction:cancel];
		[self.captureController presentViewController:confirmation animated:YES completion:nil];
	} else {
		[self hideCaptureController];
	}
}

- (void)captureScenario:(AUIImageCaptureScenario*)scenario didFailWithError:(nonnull NSError*)error
{
	NSLog(@"Capture controller error: %@", error);
	UIAlertController* alert = [UIAlertController
		alertControllerWithTitle:NSLocalizedString(@"SomethingWentWrong", nil)
		message:error.localizedDescription
		preferredStyle:UIAlertControllerStyleAlert];
	[alert addAction:[UIAlertAction
		actionWithTitle:NSLocalizedString(@"Ok", nil)
		style:UIAlertActionStyleCancel
		handler:nil]];
	[self.captureController presentViewController:alert animated:YES completion:^{
		[self.captureController setPaused:YES];
	}];
}

- (void)captureScenario:(AUIImageCaptureScenario*)captureScenario didCaptureImageWithResult:(AUIImageCaptureResult*)result
{
	NSLog(@"did capture result %@", result);
	PreviewViewController* viewController = [self.storyboard instantiateViewControllerWithIdentifier:@"Preview"];
	viewController.image = result.image;
	// Add one more image. It will be saved when user confirms
	viewController.imageIndex = self.imagePaths.count + self.tempImagePaths.count + 1;
	viewController.imagesCount = self.imagePaths.count + self.tempImagePaths.count + 1;
	viewController.delegate = self;
	UINavigationController* navigationController = [[UINavigationController alloc] initWithRootViewController:viewController];
	[self.captureController presentViewController:navigationController animated:YES completion:nil];
}

#pragma mark - Preview Controller

- (void)previewController:(PreviewViewController*)viewController didAddImage:(UIImage*)image
{
	[self saveImage:image temporary:YES];
	[self.captureController dismissViewControllerAnimated:YES completion:nil];
}

- (void)previewController:(PreviewViewController*)viewController didCompleteWithImage:(UIImage*)image
{
	[self completeCaptureSessionWithImage:image];
}

- (void)previewControllerDidCancel:(PreviewViewController*)viewController
{
	[self.captureController dismissViewControllerAnimated:YES completion:nil];
}

#pragma mark - Collection View

- (NSInteger)collectionView:(UICollectionView*)collectionView numberOfItemsInSection:(NSInteger)section
{
	return self.imagePaths.count;
}

- (UICollectionViewCell*)collectionView:(UICollectionView*)collectionView cellForItemAtIndexPath:(NSIndexPath*)indexPath
{
	CollectionViewCell* cell = [collectionView dequeueReusableCellWithReuseIdentifier:@"Cell" forIndexPath:indexPath];
	[self.imageLoader loadImageWith:self.imagePaths[indexPath.item] completion:^(UIImage* image) {
		cell.imageView.image = image;
	}];
	return cell;
}

- (void)collectionView:(UICollectionView*)collectionView didEndDisplayingCell:(UICollectionViewCell*)cell forItemAtIndexPath:(NSIndexPath*)indexPath
{
	if(indexPath.item < self.imagePaths.count) {
		NSString* path = [self.imagePaths objectAtIndex:indexPath.item];
		[self.imageLoader cancelLoading:path];
	}
}

- (CGSize)collectionView:(UICollectionView*)collectionView layout:(UICollectionViewFlowLayout*)collectionViewLayout sizeForItemAtIndexPath:(NSIndexPath*)indexPath
{
	// Calculate correct item size
	CGFloat viewWidth = collectionView.frame.size.width;
	CGFloat itemWidth = collectionViewLayout.itemSize.width;
	CGFloat space = collectionViewLayout.minimumInteritemSpacing;
	NSInteger count = viewWidth / itemWidth;
	if(count > 0) {
		itemWidth = (NSInteger)(viewWidth - space * (count - 1)) / count;
	}
	return CGSizeMake(itemWidth, itemWidth);
}

@end
